/*
 *	Functions for analysing arm architecture systems
*/

#include "../../globals.h"
#include "magic.h"

/*
 * Some alchemy:
 */

/* 
 * Function that manages mapping from linear to physical address space.
 * 
 * Good overview is available in DDI 01001.pdf (ARM documentation) at page 
 * 723+ (especially page 727)
 *
 *
 * priv_block_size from lin_addr are read using provided PGD 
 * (kernel PGD by default) into out_data pointer.
 * dump_offset is output parameter, if NULL, ignored.
 * Otherwise physical offset translated from linear address is stored here.
 * (!) Please note, that read block can be larger than one page. 
 * In this case offset is valid only for first part of data.
 */


int arm_read_linear(u_int64_t PGD, u_int64_t lin_addr,
			 void *out_data, u_int64_t priv_block_size, 
			u_int64_t *dump_offset){

	// "PGD" is needed, because each process have it's own 
	// linear address space. 
	// PGD 0 means kernel memory.
	// NOTE: Term PGD is used to refer first directory in hierarchy,

	// Find in VPT "physical" address, where data should be. 
	// (if problem, return 1)
	// Read block_size of that data into out_data 
	// (if problem, return 2)
	// return 0

	u_int64_t physical_offset = lin_addr;

	u_int64_t level1_offset = lin_addr;
	u_int64_t level2_offset = lin_addr;
	u_int64_t page_number = lin_addr;

	u_int64_t tmp_var = 0;

	dbgprint("addr: 0x%"PRIX64"",lin_addr); 

	if (PGD == 0) {	//special case
		PGD = img1.VPT_address;

	/* if VPT_address is not available,
	 * we can try "best effort", and use hard-coded 
	 * "direct mapping", usually used by OS kernels.
	 * (and OS specific)
	 * for linux on arm, mapping can be found in 
	 * Documentation/arm/memory.txt
	 * parsing here:

Start		End		Use
--------------------------------------------------------------------------
ffff8000	ffffffff	copy_user_page / clear_user_page use.
				For SA11xx and Xscale, this is used to
				setup a minicache mapping.

ffff1000	ffff7fff	Reserved.
				Platforms must not use this address range.

ffff0000	ffff0fff	CPU vector page.
				The CPU vectors are mapped here if the
				CPU supports vector relocation (control
				register V bit.)

ffc00000	fffeffff	DMA memory mapping region.  Memory returned
				by the dma_alloc_xxx functions will be
				dynamically mapped here.

ff000000	ffbfffff	Reserved for future expansion of DMA
				mapping region.

VMALLOC_END	feffffff	Free for platform use, recommended.
				VMALLOC_END must be aligned to a 2MB
				boundary.

VMALLOC_START	VMALLOC_END-1	vmalloc() / ioremap() space.
				Memory returned by vmalloc/ioremap will
				be dynamically placed in this region.
				VMALLOC_START may be based upon the value
				of the high_memory variable.

PAGE_OFFSET	high_memory-1	Kernel direct-mapped RAM region.
				This maps the platforms RAM, and typically
				maps all platform RAM in a 1:1 relationship.

TASK_SIZE	PAGE_OFFSET-1	Kernel module space
				Kernel modules inserted via insmod are
				placed here using dynamic mappings.

00001000	TASK_SIZE-1	User space mappings
				Per-thread mappings are placed here via
				the mmap() system call.

00000000	00000fff	CPU vector page / null pointer trap
				CPUs which do not support vector remapping
				place their vector page here.  NULL pointer
				dereferences by both the kernel and user
				space are also caught via this mapping.


	 * so we can try to translate PAGE_OFFSET - "high_memory-1",
	 * which should be (0xC0000000 - 0xFFFFFFFF)
  	 */  
		if (img1.VPT_not_available == 1) { 

			// kernel text direct mapping
			if ((lin_addr > 0xC0000000) &&
				(lin_addr < 0xFFFFFFFF)) {
				physical_offset=lin_addr - 0xC0000000;
				dbgprint("direct mapping 0x%"PRIX64" -> 0x%"PRIX64"",
					lin_addr, physical_offset);
			}
			// If none of mappings matches 
			if (physical_offset == lin_addr) {
				dbgprint("nothing matched, 0x%"PRIX64"",lin_addr);
				return -1;
			}

			// Read memory
			// level 0 (physical page) 
			lseek(inputfd,physical_offset,SEEK_SET); 
			dbgprint("(level 0) Dirrect mapping: Reading 0x%"PRIX64" \
				from PO: 0x%"PRIX64"\n",lin_addr,physical_offset);
			
			// save physical offset to output variable
			if (dump_offset != NULL) { 
				*dump_offset=physical_offset; 
			}

			if ( (status=read(inputfd,out_data,priv_block_size)) 
				== -1 ) {

        		        errormsg=strerror(errno);
               			fprintf(stderr, "read_liner: read() in \
					%s: %s.\n", inputfile, errormsg);
	                	return 38;
			}
			
			return 0; //another job well done.

		}	

	}

	// VMSAv6, first level descriptors (alias PGD)

	// this part (PML4E & PDPE) is same for all page sizes in long mode
	// For extracting addresses I use bitwise and,
	// for extracting offsets I have to shift bits.


	// bits ( 31:20)
	level1_offset = BS32(level1_offset,20,31);

	dbgprint("level1_offset: 0x%"PRIX64"",level1_offset);	

	dbgprint("(level 1) PGD : 0x%"PRIX64", l1_offset: 0x%"PRIX64", pointer_size: %d",
		PGD, level1_offset, img1.pointer_size);	

	// level 1
        lseek(inputfd,PGD+level1_offset*img1.pointer_size,SEEK_SET); 
	dbgprint("(level 1) reading from : 0x%"PRIX64"",
		PGD+level1_offset*img1.pointer_size);	

        if ( (status=read(inputfd,&page_number,img1.pointer_size)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "read_liner: read() in %s: %s.\n",
			inputfile, errormsg);
                return 37;
        }
	dbgprint("page_number (level 1): 0x%"PRIX64"",page_number);	

	//bits ( 1:0), determines paging table structure
	tmp_var = page_number & B64(0,1); 
	dbgprint("bits (level 1): 0x%"PRIX64"",tmp_var);	

	// Level 1 branching..

	if (tmp_var == 0) { //"fault" state
		dbgprint("Level 1 table: \"fault\"");
		return 101;
	}
	if (tmp_var == 3) { //"reserved" state (or tiny tables used)
		dbgprint("Level 1 table: \"reserved\"");
		return 102;
	}


	if (tmp_var == 1) { //"coarse" pages are used

	/* From ARM documentation:
	 * Because the upper four bits of the Page Index and 
	 * low-order four bits of the Second-level Table Index 
	 * overlap, each page table entry for a large page must
	 * be repeated 16 times (in consecutive memory locations)
	 * in a coarse page table.
	 */
		dbgprint("coarse pages used for this page.");
		// bits 19:12
	        level2_offset = BS32(level2_offset,12,19);
	
		//bits ( 31:10), but no shift. Offset will be added later
		page_number = page_number & B32(10,31);

		dbgprint("(level 2) Reading 0x%"PRIX64" from PO: 0x%"PRIX64"\n",
			lin_addr,page_number+level2_offset);

		//level 2 
		lseek(inputfd,page_number+level2_offset,SEEK_SET);
	        if ( (status=
			read(inputfd,&page_number,img1.pointer_size)
			) == -1 ) {

			errormsg=strerror(errno);
			fprintf(stderr, "read_liner: read() in %s: %s.\n",
				inputfile,errormsg);
			return 37;
		}
		dbgprint("page_number (level 2): 0x%"PRIX64"",page_number);	

//345678901234567890123456789012345678901234567890123456789012345678901234567890
		//bits ( 1:0), determines paging table structure
		tmp_var = page_number & B64(0,1);
		dbgprint("bits (level 1): 0x%"PRIX64"",tmp_var);	

		if (tmp_var == 0) { // "falt"
			dbgprint("Level 2 table: \"fault\"");
			return 103;
		}
		if (tmp_var == 1) { //"large" (64KB) pages are used
			dbgprint("\"large\" pages used for this page.");
		
			// address of physical page
			// bits 31:16
			page_number = page_number & B64(16,31);

			//offset in page, bits ( 15:0). 
			physical_offset = physical_offset & B64(0,15);

			dbgprint("(level physical) Reading 0x%"PRIX64" from PO: \
				0x%"PRIX64"\n", lin_addr, 
				page_number+physical_offset);

			// physical level
			lseek(inputfd,page_number+physical_offset,SEEK_SET); 
			if (dump_offset != NULL) { 
				//save physical offset to output variable
				*dump_offset=page_number+physical_offset;
			}

			// If we want to read more than 1 page: use recursion
		
			// offset + block_size must be less then 64KB
			if (priv_block_size > (KB_64-physical_offset)){ 
				dbgprint("arm_read_linear: page too small for \
					reading... processing another one.");
				// recursion	
				tmp_var = KB_64-physical_offset;

				// Process rest of request recursively... 
				// beware about effectives
				arm_read_linear(PGD, lin_addr+tmp_var,
						out_data+tmp_var,
						priv_block_size-tmp_var, NULL); 
			} else{
				tmp_var=priv_block_size;
			}

			if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
				errormsg=strerror(errno);
				fprintf(stderr, "read_liner: read() in %s: \
					%s.\n",inputfile,errormsg);
				return 38;
			}

		}

		if ((tmp_var == 2) || (tmp_var == 3)) { 
			// "small" or "extended small" (4KB) pages are used 
			// (only if subpages are enabled,
			//  otherwise "extended small")
			// for our purpose, this is not important,
			// offsets are same
			dbgprint("\"small\" pages used for this page.");
		
			// address of physical page
			// bits 31:12
			page_number = page_number & B64(12,31);

			// offset in page, bits ( 11:0). 
			physical_offset = physical_offset & B64(0,11);

			dbgprint("(level physical) Reading 0x%"PRIX64" from PO: \
				0x%"PRIX64"\n",lin_addr,
				page_number+physical_offset);

			// physical level
			lseek(inputfd,page_number+physical_offset,SEEK_SET); 
			if (dump_offset != NULL) { 
				//save physical offset to output variable
				*dump_offset=page_number+physical_offset;
			}

			// If we want to read more than 1 page: use recursion
		
			// offset + block_size must be less then 64KB
			if (priv_block_size > (KB_4-physical_offset)){ 
				dbgprint("arm_read_linear: page too small for \
					reading... processing another one.");
				// recursion	
				tmp_var = KB_4-physical_offset;

				// Process rest of request recursively... 
				// beware about effectives
				arm_read_linear(PGD, lin_addr+tmp_var,
						out_data+tmp_var,
						priv_block_size-tmp_var, NULL);
			} else{
				tmp_var=priv_block_size;
			}

			if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
				errormsg=strerror(errno);
				fprintf(stderr, "read_liner: read() in %s: \
					%s.\n",inputfile,errormsg);
				return 38;
			}
		} 
	}
	
	if (tmp_var == 2) { 
		// "section" (or "supersection") pages are used

		dbgprint("section pages used for this page.");
		//TODO 16MB supersection support

		// address of physical page
		// bits 31:20
		page_number = page_number & B64(20,31);

		// offset in page, bits ( 19:0). 
		physical_offset = physical_offset & B64(0,19);

		dbgprint("(level physical) Reading 0x%"PRIX64" from PO: 0x%"PRIX64"\n",
			lin_addr, page_number+physical_offset);

		// physical level
		lseek(inputfd,page_number+physical_offset,SEEK_SET);
		if (dump_offset != NULL) { 
			// save physical offset to output variable
			*dump_offset=page_number+physical_offset;
		}

		// If we want to read more than 1 page: use recursion
		
		// offset + block_size must be less then 1MB
		if (priv_block_size > (MB_1-physical_offset)){ 
			dbgprint("arm_read_linear: page too small for reading\
				... processing another one.");

			// recursion	
			tmp_var = MB_1-physical_offset;
			// Process rest of request recursively... 
			// beware about effectives
			arm_read_linear(PGD, lin_addr+tmp_var, out_data+tmp_var,
					priv_block_size-tmp_var,NULL); 
		} else{
			tmp_var=priv_block_size;
		}

		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
			errormsg=strerror(errno);
			fprintf(stderr, "read_liner: read() in %s: %s.\n",
				inputfile,errormsg);
			return 38;
		}

	}

        return 0;

}


/*
 * Function for verifying if given address contains system VPT 
 * (resp. PaGeDirectory)
 *
 * Verification for arm is done in few steps:
 * - Address must be 16KB alligned (arm specification)
 * - for each entry in table, type is determined
 * - if "course" page, then bits 2,3,4 ShouldBeZero
 * - if "section" page, then at least bit 19 ShouldBeZero
 *
 * During whole check errors and records are counted.
 * - If more then ARM_VPT_MAGIC_ERR_FATAL errors is found,
 * check is terminated immediately 
 * - There must be at least ARM_VPT_MIN_REC records in table
 * - The ratio of records and errors must be less then
 *   ARM_VPT_MAGIC_ERR_RATIO
 * 
 * In practical tests, ratio condition seems to be 
 * very good filter. 
 *
 * arm_VPT_test (physical address to verify )
 * return 0 on success, otherwise error > 0
 *
 */

int arm_VPT_test(u_int64_t my_addr){

	u_int64_t tmp_var;
	int mi,ecounter=0,rcounter=0;

	// using 32bit variables. (!)
	u_int32_t small_page[ARM_VPT_MAGIC_BS/ARM_POINTER_SIZE];

	// must be 16KB aligned

	tmp_var = my_addr & (u_int64_t)(KB_16-1);
	
	dbgprint("0x%"PRIX64"",my_addr);

	if (tmp_var != 0) {
		dbgprint("Not 16KB aligned!");
		return 1;
	}

	lseek(inputfd,my_addr,SEEK_SET); 

	// read whole page (16KB)
	if ( (status=read(inputfd,&small_page,ARM_VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "arm_VPT_test: read() in %s: %s.\n",
			inputfile,errormsg);
                return 99;
        }
	
	for (mi=0;mi < (ARM_VPT_MAGIC_BS/ARM_POINTER_SIZE);mi++) { 
		// page

		// Last 2 bits determines type of page entry
		// section
		if ((small_page[mi] & B32(0,1)) == 1) { 
			// course page entry
			// => bits 2,3,4 SBZ (should be zero)
			rcounter++;
			if ((small_page[mi] & B32(2,4)) != 0) {
				dbgprint("err: at %d is 0x%X", mi, 
					small_page[mi]);
				if (ecounter++ > ARM_VPT_MAGIC_ERR_FATAL) {
					dbgprint(" addr: 0x%"PRIX64" mi=%d, \
						ecounter: %d, rcounter: %d\n",
						my_addr,mi,ecounter,rcounter);

					// tolerate small amount of errors
					return 2; 			
				}
			}
		}	
		if ((small_page[mi] & B32(0,1)) == 2) { 
			// section page entry
			// => at least bit 19 SBZ (should be zero)
			// in some modes bits 19-15, but 19 enough for us)
			rcounter++;
			if ((small_page[mi] & B32(19,19)) != 0) {
				dbgprint("err: at %d is 0x%X", mi,
					 small_page[mi]);
				if (ecounter++ > ARM_VPT_MAGIC_ERR_FATAL) {
					dbgprint(" addr: 0x%"PRIX64" mi=%d, \
						ecounter: %d, rcounter: %d\n",
						my_addr,mi,ecounter,rcounter);
					// tolerate small amount of errors
					return 3; 
				}

			}
		}
	}	


	if (rcounter < ARM_VPT_MIN_REC) {
		dbgprint(" addr: 0x%"PRIX64" mi=%d, ecounter: %d, rcounter: %d\n",
			my_addr,mi,ecounter,rcounter);
		return 4; // require at least few valid records
	}

	// Additional check: 
	// errors/valid records must pass some ratio
	// Direct seeking for errors is too strict for full tables

	if (ecounter != 0) { //division by zero
		if ( (rcounter/ecounter) < ARM_VPT_MAGIC_ERR_RATIO) {
			dbgprint(" addr: 0x%"PRIX64" mi=%d, ecounter: %d, \
				rcounter: %d",my_addr,mi,ecounter,rcounter);
			return 5; // require low error ratio
		}
	}
	
	return 0;
}

/*
 * Function for verifying if given virtual address contains VPT 
 * (resp. PaGeDirectory) of some process.
 * 
 * Difference between user_VPT_test and VPT_test is that
 * user_VPT_test take virtual address as input parameter.
 * Also on some architectures, this can be completely different 
 * heuristics.
 * 
 */

int arm_user_VPT_test(u_int64_t my_addr){

	int status=0;
	u_int64_t empty=0,phys_addr=0;

	dbgprint("0x%"PRIX64"",my_addr);

	// Translate address from virtual to physical..
	arm_read_linear(0,my_addr,&empty,img1.pointer_size,&phys_addr); 
	
	// Test
	if ((status=arm_VPT_test(phys_addr))!=0) {	
		dbgprint("arm_VPT_test returned %d.",status);
		return 1;
	}

	return 0;
}


/*
 * Function that tries to determine general information 
 * about analyzed image.
 *
 * for now, only system VPT is found.
 * if mode == 1, then all possible VPT candidates are printed.
 *
 * arm_find_VPT ()
 * return 0 on success, otherwise error > 0
 * Discovered information are saved in global structure img1.
 *
 */


int arm_find_VPT(int mode){

	int status=1;
	u_int64_t addr=ARM_VPT_MAGIC_ADDR;
	
	// I realised, that I don't get too much information in 
	// this phase of analysis 
	
	// Special mode, only find all possible VPT candidates
	if (mode == 1) {
		printf("Listing VPT candidates(arm):\n");
		while ( (addr + ARM_VPT_MAGIC_BS) <= img1.ram_size) {
			if (arm_VPT_test(addr) == 0) {
				printf("- Possible VPT at: 0x%"PRIX64"\n",
					addr);
			}
			addr += ARM_VPT_MAGIC_BS; // 16KB
			dbgprint("step, addr: 0x%"PRIX64"\n",addr);
		}

		return 0;	
	} 

	// Otherwise normal approach


//	printf("arm_find_VPT:\n");
	
	if (img1.VPT_user_specified == 0 ) {	
		// if image address was not manually specified
		while (addr < ( ARM_VPT_MAGIC_ADDR + 
			(ARM_VPT_MAGIC_BS * ARM_VPT_MAGIC_COUNT))) {

			if ((status=arm_VPT_test(addr))==0) { // test addr
				img1.VPT_address=addr; // found one! :)
				printf("VPT address found at: 0x%"PRIX64"\n",
					img1.VPT_address);
				return 0;
			}

			addr += ARM_VPT_MAGIC_BS; // 16KB
			dbgprint("step, addr: 0x%"PRIX64"\n",addr);
		}
		// we have tried too many pages, 
		// so something is wrong probably
		// further search will be useless
		printf("ARM VPT_address not found, terminating at: 0x%"PRIX64"\n",
			addr);
		return 1;	
	} else {
		if (img1.VPT_not_available == 0 ) {	
			// Use of PGD is not disabled	
			printf("User specified VPT address: 0x%"PRIX64"\n",
				img1.VPT_address);
		} else {
			printf("Using hardcoded mappings instead of VPT.\n");
			printf("This may produce strange results!\n");
		}
	}

	return 0;  // 0 if all OK, otherwise error
}

/*
 * Function that guess, if we are really 
 * analysing arm architecture.
 *
 * return 0 on success, otherwise error > 0
 *
 */

int arm_guess(){
	return 1;
}

/*
 * Function that sets pointers to all other 
 * architecture dependent functions and hooks
 *
 * arm_arch_settings()
 * return 0 on success 
 */

int arm_arch_settings(){

	strncpy(img1.architecture,"arm\0",4);
	img1.read_linear=(int (*)())arm_read_linear;
	img1.VPT_test=(int (*)())arm_VPT_test;
	img1.user_VPT_test=(int (*)())arm_user_VPT_test;
	img1.find_VPT=(int (*)())arm_find_VPT;
	img1.pointer_size=ARM_POINTER_SIZE;
	img1.integer_size=ARM_INTEGER_SIZE;

	return 0;
}

int arm_init(struct arch_struct *self){

	strncpy(self->architecture,"arm\0",4);
	self->read_linear=(int (*)())arm_read_linear;
	self->VPT_test=(int (*)())arm_VPT_test;
	self->user_VPT_test=(int (*)())arm_user_VPT_test;
	self->find_VPT=(int (*)())arm_find_VPT;
	self->guess=(int (*)())arm_guess;
	self->pointer_size=ARM_POINTER_SIZE;
	self->integer_size=ARM_INTEGER_SIZE;

	return 0;
}



