/*
 *	Functions same for analysing all architectures
*/

#include "../globals.h"
#include "i386/functions_i386.h"
#include "x86_64/functions_x86_64.h"
#include "arm/functions_arm.h"


/*
 * Some alchemy:
 */


	
/*
 * Function that sets architecture specific 
 * settings according to values in supplied structure
 *
 * arch_settings (pointer to architecture struct)
 * return 0 on success, otherwise error > 0
 * results are stored in global structure
 *
 */

int arch_settings(struct arch_struct *arch_to_set){

	strncpy(img1.architecture,arch_to_set->architecture,ARCH_NAME_LEN);
	img1.read_linear=arch_to_set->read_linear;
	img1.VPT_test=arch_to_set->VPT_test;
	img1.user_VPT_test=arch_to_set->user_VPT_test;
	img1.find_VPT=arch_to_set->find_VPT;
	img1.pointer_size=arch_to_set->pointer_size;
	img1.integer_size=arch_to_set->integer_size;

	return 0;
}

/*
 * Function that sets pointer to correct functions,
 * according to detected architecture
 * This must be done by checking string;(
 * Index in array can change.
 *
 * set_arch_from_list (string with architecture name)
 * return 0 on success, otherwise error > 0
 * results are stored in global structure
 *
 */


int set_arch_from_list(char *arch_name){
	int status,j;
	for (status=0,j=1;j<=img1.archcount;j++) {
		// Compare user specified arch and one of list		
		status=strncmp(architectures[j].architecture,
			arch_name,
			ARCH_NAME_LEN);

		if (status == 0) {
			dbgprint("Setting arch: %s (%d)",
				arch_name,j);
			arch_settings(&architectures[j]);
			img1.arch=j; //set index
			return 0;
		}
	}
	return 1;
}

/*
 * Function with heuristic for architecture guessing
 *
 * This function is not really useful.
 * Original idea was that there may exists some 
 * heuristic, that can identify different types 
 * of architectures. I was not able to find such heuristic.
 * Because number of different architectures is small,
 * best solution seem to be try then one by one.
 * Usually, this is not needed, because user 
 * knows architecture of dumped system. (or have a hint) 
 *
 *
 * guess_architecture ()
 * return 0 on success, otherwise error > 0
 * results are stored in global structure
 *
 */

int guess_architecture(){

	int i, status=0;

	printf("Guessing architecture: ");

	// Try heuristics for all architectures.
	// First architecture that match is set

	for (i = 1; i <= img1.archcount;i++) {
		dbgprint("%d",i);
		status = architectures[i].guess();
		if (status == 0) {
			dbgprint("%d: %d",i,status);
			printf("OK, %s.\n",
				architectures[i].architecture);
			// set pointers and other settings
			img1.arch = i; 
			arch_settings(&architectures[i]);

			return 0;
		}
	}

	// If none match...
	// Stupid failsafe	
	
	// Use first available (index 1)
	printf("Fail.\nUsing first available architecture: %s.\n",
		architectures[1].architecture);
	// set pointers and other settings
	img1.arch=1; 
	arch_settings(&architectures[1]);

	return 0;
}

/*
 * Function that goes trough simple list and counts its size
 *
 * Simple list look like this:
 *
 *     entry    entry    entry     entry
 *       |    --> |    --> |    --> ...
 *   off |   --   |   --   |   --      
 *   set |  --    |  --    |  --       
 *       |---     |---     |---        
 *
 * offset is distance between address where pointer points
 * and address of next pointer 
 *
 * simple_list_size(address , offset in structure)
 * return size of list on success, otherwise error < 0
 *
 */

int simple_list_size(u_int64_t addr,int offset){

	u_int64_t old=0,next=0;
	int size=0;

	dbgprint("addr: 0x%"PRIX64", offset %d",addr,offset);
	for (old=addr;;size++){
		dbgprint("ptr: 0x%"PRIX64", size %d",old,size);
	
		if (size > MAX_LIST_SIZE ) {
			return -2; //List too long
		} 	

		status=img1.read_linear(0, old+offset, &next,
					img1.pointer_size, NULL);
		if ( status ) {
			errormsg=strerror(errno);
			dbgprint("read() in %s: %s.\n",inputfile,errormsg);
			return -1; //List should be ended by NULL pointer.
		}
	
		if ((old == next ) || (next == 0 )) {
			return (size+1); 
			// It is strange, but sometimes list 
			// seems to be finished by pointer to itself.
		}
		old=next;
	}

}

/*
 * Function for verifying if given address contains structure list_head.
 *
 * Verification is done by finding and checking double linked 
 * "list_head". This way is chosen instead of classic 
 * pattern matching, because I believe it is more flexible and 
 * can be used for almost every 2.6 kernel version. (even future;)
 * It does not depend on task_struct definition too much, only 
 * important characteristic is number of list_heads between 
 * "comm" array and list_head "tasks". This is specified by
 * TS_MAGIC_HEADLIST_NUMBER. 
 *
 * Correct list_head can be verified easily:
 * Data at address point to (virtual) address where is stored another
 * list_head. Second pointer in this second head list point back
 * to (virtual) address where we started.
 *
 *  ...   |--------> |---------->  |---------->   ...
 *          <------  | <-----------| <-------- 
 *
 * Note: This will not find processes, that are inactive,
 * has already ended in past, or something similar. 
 * Therefore, if you want to discover all the processes, 
 * you have to generate pattern and then use bgrep utility.
 *
 *
 * verify_cyclic_list_head (address to verify )
 * return 0 on success, otherwise error > 0
 *
 */

int verify_cyclic_list_head(u_int64_t ver_addr){

	u_int64_t next=0,back=0,next_from_back=0;

	dbgprint("start, addr: %"PRIX64"",ver_addr); 
	if ( (status=lseek(inputfd,ver_addr,SEEK_SET)) == -1 ) {
		errormsg = strerror(errno);
		dbgprint("lseek() to 0x%"PRIX64" "\
			"failed because of %s.\n", ver_addr, errormsg);
		return(1);
	}

	if ( (status=read(inputfd,&next,img1.pointer_size)) == -1 ) {
		errormsg=strerror(errno);
		dbgprint("read() in %s: %s.\n",
			inputfile, errormsg);
		return(2);
	}

	if (next == 0) {
		dbgprint("Zero.\n");
		return 6;
	}

	status = img1.read_linear(0,next,&back,img1.pointer_size,NULL);
	if ( status == -1 ) {
		dbgprint("0x%"PRIX64" seems is not "\
			"valid virtual address.\n", next);
		return(3);
	}
	status = img1.read_linear(0, 
				back+img1.pointer_size, 
				&next_from_back,
				img1.pointer_size, 
				NULL);
	if ( status == -1 ) {
		dbgprint("0x%"PRIX64" seems is not "\
			"valid virtual address.\n", next_from_back);
		return(4);
	}
	
	if (next != next_from_back) {
		dbgprint("addresses 0x%"PRIX64" and "\
			"0x%"PRIX64" does not match.\n",next,next_from_back);
		return(5);
	}
	
	// We need to verify second pointer of list_head too, 
	// there were some false positives
	// same algorithm, except that "prev" pointer points to 
	// (address - img1.pointer_size) of reverse pointer.
	
	status=lseek(inputfd,ver_addr-img1.pointer_size,SEEK_SET);
	if ( status == -1 ) {
		errormsg = strerror(errno);
		dbgprint("2: lseek() to 0x%"PRIX64" "\
			"failed because of %s.\n",ver_addr,errormsg);
		return(6);
	}

	if ( (status=read(inputfd,&next,img1.pointer_size)) == -1 ) {
		errormsg=strerror(errno);
		dbgprint("2: read() in %s: %s.\n",
			inputfile, errormsg);
		return(7);
	}

	if (next == 0) {
		dbgprint("Zero.\n");
		return 8;
	}

	status = img1.read_linear(0, next+img1.pointer_size,
				&back, img1.pointer_size, NULL);
	if ( status == -1 ) {
		dbgprint("2: 0x%"PRIX64" seems is not "\
			"valid virtual address.\n", next);
		return(9);
	}
	
	status = img1.read_linear(0, back, &next_from_back,
					img1.pointer_size, NULL);
	if ( status == -1 ) {
		dbgprint("2: 0x%"PRIX64" seems is not "\
			 "valid virtual address.\n", next_from_back);
		return(10);
	}
	
	if (next != next_from_back) {
		dbgprint("2: addresses 0x%"PRIX64" and "\
			"0x%"PRIX64" does not match.\n", next, next_from_back);
		return(11);
	}

	return 0;
}



/*
 *  lin_verify_vfsmount (pointer to vfsmount structure)
 *  return 0 on success, otherwise -1
 *
 *  vfsmount structure is "recognized" by simple test:
 *  should contain pointer to string "/dev/".
 *  (char *mnt_name)
 *  This is alone not a strong heuristic, but in combination
 *  with specific chain where structure should be found when trying 
 *  to list open files, it is enough.
 */

int lin_verify_vfsmount(u_int64_t ver_addr) {

	return -1;
}



/*
 *  lin_verify_qstr (pointer to "qstr" structure )
 *  return 0 on success, otherwise error
 *
 *  Used for listing of open files per process.
 *    d_name -> name
 *    (qstr)   (char)
 *
 *  This function verifies, if pointer points to qstr structure.
 *  qstr is elemental structure, and we assume, 
 *  that will not change (there is no reason)
 *  Verification is done by checking, 
 *  if string that "name" points to is really "len"
 *  bytes long (looking for \0) and "len" is not 0.
 *
 *  This is qstr definition:
 *
 *  struct qstr {
 *          unsigned int hash;
 *          unsigned int len;
 *          const unsigned char *name;
 *  };
 *                          
 *
 */

int lin_verify_qstr(u_int64_t ver_addr) {
	
	return 0;
}

/*
 *  lin_verify_dentry (pointer to "dentry" structure )
 *  return 0 on success, otherwise -1
 *  determined distances are saved into img1
 *
 *  Used for listing of open files per process.
 *  Chain we are following looks like:
 *
 *  => task_struct ->       files   ->      fdt     ->      fd      ->      f_path  ->      dentry  ->      d_name-> name
 *  		    (files_struct)      (fdtable)        (file **)          (path)          (dentry)        (qstr)  (char)
 *
 *  This function determines distances in structure dentry
 *  It is better to determine structures by more "sub" heuristics,
 *   because amount of tries needed is decreased.
 *  (and amount of false positives too)
 *  
 */

int lin_verify_dentry(u_int64_t ver_addr) {

	return 0;
}

/*
 *  lin_determine_up_to_path (pointer to "tasks" list_head in task_struct )
 *  return 0 on success, otherwise -1
 *  determined distances are saved into img1
 *
 *  function that tries to determine distances in structures needed
 *  for listing of open files per process up to f_path structure.
 *  Chain we are following looks like:
 *
 *  => task_struct ->       files   ->      fdt     ->      fd      ->      f_path  ->      dentry  ->      d_name-> name
 *  		    (files_struct)      (fdtable)        (file **)          (path)          (dentry)        (qstr)  (char)
 *
 *  So this function determines distances in structures 
 *  files_struct, fdtable, file and path.
 *  It is better to determine structures by more "sub" heuristics,
 *  because amount of tries needed is decreased.
 *  (and amount of false positives too)
 *  
 */

int lin_determine_up_to_path(u_int64_t ver_addr) {

	return -1;	// Not found
}

/*
 *  determine_open_files (pointer to "tasks" list_head in task_struct )
 *  return 0 on success, otherwise -1
 *
 *  function that tries to determine distances in all structures needed
 *  for listing of open files per process. 
 *  This will definitely not work in 2.4 kernels.
 */

int lin_determine_open_files(u_int64_t ver_addr) {

		
	return -1;
}

/*
 *  function that tries to determine distances in all structures needed
 *  for listing of open files per process. 
 *  Should work in most BSD kernels.
 *
 *  => struct proc ->     p_fd     ->   fd_ofiles   ->   TODO
 *  		      (filedesc*)        (file**)        ()
 *
 *
 *  bsd_determine_open_files (pointer to "proc" list_head in proc struct )
 *  return 0 on success, otherwise -1
 * 
 */

int bsd_determine_open_files(u_int64_t ver_addr) {

	// TODO?
	// there is no standard lsof utility in bsd.
	// listing file names seems to be quite complicated this way.

	dbgprint("NOT IMPLEMENTED");
		
	return -1;
}

/*
 *  verify_task_struct (physical address of string with process 
 *			name inside process structure )
 *  
 *  Function verify, if there are some list_heads (cyclic or bsd style)
 *  around given address. List_heads must be long enough.
 *
 *  Works for multiple operating systems, (lin, win, bsd)
 *
 *  return 0 on success, otherwise negative value
 */

int verify_task_struct(u_int64_t ver_addr) {

	int i; 			// for for;)
	int status; 		
	int hl_index=-1;	// headlist index in array
	int hl_to_try;		// how many hl we can find
	int spread;		// determining right hl
	int stop_spread;	// determining right hl
	int hl_size;		// size of found hl

	// size of strings in found hl
	int max_bsd_list_strings_size=0;

	// "start" pointers on list
	u_int64_t list_start=0,bsd_list_start=0;
	
	// if OS is not known, we can try to determine it
	int found_os=0;
				

	dbgprint("0x%"PRIX64"\n",ver_addr);
	// initialize	
	hl_to_try=img1.os1->ts_magic_headlist_max;

	// look for list_heads around string with name
	for (i = 1;i < (img1.os1->ts_magic_max_tries);i++){
		
		dbgprint(" %d: %"PRIX64"",i, ver_addr - (i * img1.pointer_size));
		status=verify_cyclic_list_head(
				ver_addr - ( i * img1.pointer_size) );
		
		if (status == 0) {	
			// found some list_head
			dbgprint("Found some list_head at negative offset %d",
				i*img1.pointer_size);
			//save each list_head
			img1.ts_headlists[++hl_index] =
				ver_addr - (i * img1.pointer_size);

			dbgprint("headlists[%d]:%"PRIX64" \n",
					hl_index,
					img1.ts_headlists[hl_index]);

			i++; 
			/* because of there can be (and are) 3 head lists 
			   in row, and without this
			   algorithm would find 5 "list_heads" */

			if (--hl_to_try == 0 ) { 
				//maximum list_head's found
				// this should not happen
				dbgprint("maximum list_head's found: %d",
					TS_MAGIC_HEADLIST_MAX);
				break; // enough head list found
			}		
		}
	}
	dbgprint("listheads found: %d, needed: %d\n",
		hl_index, 
		img1.os1->ts_magic_headlist_number);


	if (hl_index < img1.os1->ts_magic_headlist_number) { 
		// It is not task_struct
		return -1;
	}

	spread = hl_index - img1.os1->ts_magic_headlist_front_number;

	dbgprint("spread: %d\n",spread);

	// This step is needed, 
	// because verify_cyclic_list_head seems to produce 
	// some false positives, 
	// so TS_MAGIC_HEADLIST_FRONT_NUMBER may be not enough accurate.

	hl_size = 0;
	i = max(spread - img1.os1->ts_magic_headlist_spread, 0);
	stop_spread = min(spread + img1.os1->ts_magic_headlist_spread,
			 hl_index);

	for (;i <= stop_spread; i++){ 

		dbgprint("listing: 0x%"PRIX64"",img1.ts_headlists[i]);

		// HOOK
		dbgprint("before hook");
		// Call OS specific hook
		if (img1.os1->hook_verify_task_struct(
					&hl_size, 
					i,
					ver_addr,
					&bsd_list_start,
					&found_os,
					&max_bsd_list_strings_size) != 0) {
			// not process structure
			return -2;
		} 
		dbgprint("after hook");
	
	}

	// Additional check, if found list_head is not too short
	if ( hl_size < img1.os1->ts_magic_min_list_size) {
		dbgprint("suspicious struct at %"PRIX64", but too \
				short list (%d).\n",ver_addr,hl_size);
		dbgprint("suspicious struct at %"PRIX64", but too \
				short list (%d).\n",ver_addr,hl_size);
		return -8;
	}

	// So at this point we know position of right list_head

	printf("- verify_task_struct: List size %d\n",hl_size);
	dbgprint("Found tasks at negative offset %d\n",i*img1.pointer_size);

	// set distance from list_head to process name string for this kernel.
	img1.ts_hl_comm_dst = ver_addr - 
			img1.ts_headlists[img1.ts_headlists_number] +
			img1.pointer_size;

	if ( (status=lseek(inputfd,
			img1.ts_headlists[img1.ts_headlists_number],
			SEEK_SET)) == -1 ) {

		errormsg = strerror(errno);
		fprintf(stderr, "	verify_task_struct: lseek() "\
			"to 0x%"PRIX64" failed because of %s.\n",
			ver_addr, errormsg);
		return(-2);
	}

	if ( (status=read(inputfd, &list_start, img1.pointer_size)) == -1 ) {
		errormsg=strerror(errno);
		fprintf(stderr, "	verify_cyclic_list_head: read() "\
				"in %s: %s.\n", inputfile, errormsg);
		return(-3);
	}

	dbgprint("Adding process at 0x%"PRIX64" to img1.ts_one_of_lists.",
		list_start);

 
	// HOOK
	dbgprint("before hook2");
	// Call OS specific hook
	if (img1.os1->hook_verify_task_struct2(
				found_os,
				bsd_list_start,
				&list_start) != 0){
		// something is wrong
		return -2;
	} 
	dbgprint("after hook2");

	img1.ts_one_of_lists=list_start;
	return 0;  // Found => It is task_list

}

/*
 *
 *  On linux, interesting fields in "module" are just near. 
 *  This should not change (no reason), 
 *  so this function is much more simple then verify_task_struct.
 *  BSD system seems to use only few (if even some) modules at runtime,
 *  so this type of heuristic would not work anyway.
 *  More simplicity means better reliability and failure resistance.
 *  Modify this function, if you are dealing with some exotic OS.
 *
 *  lin_verify_module (pointer to "name" field of module )
 *  return distance between name and list_head in bytes on success, 
 *  otherwise < 0
 *  
 */

int lin_verify_module(u_int64_t ver_addr) {

	int status,size;
	u_int64_t one_of=0;

	dbgprint("verify_module: %"PRIX64": %"PRIX64" ::\n",
		ver_addr, ver_addr-(img1.pointer_size)); 
	status = verify_cyclic_list_head(ver_addr - (img1.pointer_size)); 
		
	if (status != 0) {		
		// no list_head found
		return -1;
	}
	
	// otherwise found some list_head



	dbgprint("Found some list_head at negative offset %d\n",
		img1.pointer_size); 

	//save list_head
	img1.md_headlists[0]=ver_addr-(img1.pointer_size); 
	dbgprint("headlists[%d]:%"PRIX64" \n",status,img1.ts_headlists[0]);

	//If this is really module

	size=lin_list_head_size(img1.ts_headlists[0]);
	dbgprint("size (size): %d of %"PRIX64"\n",size,img1.md_headlists[0]);
	if (size > 0) {// more and not loop (loop < 0)
		//save which list_head is right one
		img1.md_headlists_number=0; 	
	}


	// So at this point we know position of right list_head

	printf("- lin_verify_module: Found module struct at "\
			"negative offset %d\n", 
			img1.pointer_size+img1.integer_size); 
	// + integer size because of enum

	//set distance from headlist to name for this kernel.
	img1.md_hl_name_dst = ver_addr - img1.md_headlists[0] 
		+ img1.pointer_size; 

	status=lseek(inputfd,img1.md_headlists[0],SEEK_SET);
	if ( (status) == -1 ) {
		errormsg = strerror(errno);
		fprintf(stderr, "lin_verify_module: lseek() to "\
				"0x%"PRIX64" failed because of %s.\n",
				ver_addr, errormsg);
		return(-2);
	}

	if ( (status=read(inputfd,&one_of,img1.pointer_size)) == -1 ) {
		errormsg=strerror(errno);
		fprintf(stderr, "lin_verify_module: "\
				"read() in %s: %s.\n", inputfile, errormsg);
		return(-3);
	}

	dbgprint("img1.md_one_of_lists: 0x%"PRIX64"",one_of);

	//hm, dirty trick, this should be in determine module
	img1.md_one_of_lists=one_of; 
	return 0;  // Found => It is module
}


/* 
 * Function for finding VPT of given process 
 * (resp. PaGeDirectory).
 *
 * This function is assumed to be called only once.
 * When distances used in analyzed image are known,
 * it is faster and safer to use them directly.
 * For each architecture appropriate heuristic function is called.
 *
 * lin_find_process_VPT (linear address of "tasks" list_head)
 * return 0 on success, otherwise error > 0
 * Discovered distances are saved in global structure img1.
 *
 * NOTE: ts_hl_mm_dst distance must be known before calling this function.
 * 
 */

int lin_find_process_VPT(u_int64_t my_addr){

	int mi=0;
	u_int64_t mm_struct=0,adept=0;
	int tries;	// how many time to try

	dbgprint("0x%"PRIX64"",my_addr);
	//read mm_struct address
	status=img1.read_linear(0, my_addr+img1.ts_hl_mm_dst, &mm_struct,
				img1.pointer_size, NULL);
	if (status) { 
		errormsg=strerror(errno);
		fprintf(stderr, "lin_find_process_VPT: read() in %s: %s.\n",
			inputfile, errormsg);
		return -2; //bad ptr
	}

	tries= img1.os1->mm_magic_pgd_dst_p*img1.pointer_size;
	//guessing "mm_pgd_dst"
	for (mi=0;mi < tries;mi += img1.pointer_size) {

		status=img1.read_linear(0, mm_struct+mi, &adept,
					img1.pointer_size, 0);
		if ( status ) {
			errormsg=strerror(errno);
			fprintf(stderr, "lin_find_process_VPT: read() "\
				"in %s: %s.\n", inputfile, errormsg);
		}
		
		dbgprint("mi: %d, read from 0x%"PRIX64", read 0x%"PRIX64"",
			mi, mm_struct+mi, adept);

		// On all supported architectures, 
		// PGD address is aligned at least on 1KB boundary. 
		// This simple check increases reliability.
		if ( (adept & 0xFFF) == 0  && (adept != 0)) {
			// if this pointer points to VPT (resp. PGD)
			// virtual address is tested
			if ( img1.user_VPT_test(adept) == 0) { 
				// then set distance we found 
				// (inside mm_struct from start to pgd)
				img1.mm_pgd_dst=mi; 
				printf("Found userspace PGD at 0x%"PRIX64"\n",
					adept);
				printf("Determined mm_struct pgd offset as %d\n",
					mi);
				return 0;
			}
		}
		debugoff;
	}
	return 1; //did not found VPT(resp. PGD)
}


/*
 * Function for finding VMAs of given process 
 *
 * This function is assumed to be called only once,
 * after "ts_hl_mm_dst" is known.
 * When distances used in analyzed image are known,
 * it is faster and safer to use them directly.
 *
 * VMAs are parts of process mapped memory,
 * we need their addresses to perform full dump
 * of process memory.
 * 
 * Function tries address around values that was used in 
 * available linux 2.6 kernel, and tries, if they match definition.
 * Success is not guaranteed, but it was working nice even for kernels
 * with debugging symbols compiled in.
 *
 * lin_find_process_VMA (linear address of "tasks" list_head)
 * return 0 on success, otherwise error > 0
 * Discovered distances are saved in global structure img1.
 *
 * */

int lin_find_process_VMA(u_int64_t my_addr){

	u_int64_t mm_struct=0,vm_area=0,mm_back=0;
	int i=0,max=0;
	int status;	
	int tries;	// how many times to try

	dbgprint("address: 0x%"PRIX64"",my_addr);

	tries=((img1.os1->ts_magic_hl_mm_dst_p) * img1.pointer_size);

	for (;i < tries; i += img1.pointer_size) {

		dbgprint("reading from 0x%"PRIX64", offset: %d", my_addr+i,i);
		//first read pointer to mm_struct
		status=img1.read_linear(0, my_addr+i, &mm_struct,
					img1.pointer_size, NULL);
		if (status) { 
			errormsg=strerror(errno);
			dbgprint("read() in %s: %s.\n", inputfile, errormsg);
		}

		dbgprint("read mm_struct: 0x%"PRIX64"",mm_struct);
	
		// read first member of mm_struct structure
		// this should be vm_area_struct *
		// If we want to be more generic, (other OSes)
		// offset determination should be performed.
		status=img1.read_linear(0, mm_struct, &vm_area,
					img1.pointer_size, NULL); 

		if (status) {	
			errormsg=strerror(errno);
			dbgprint("read() in %s: %s.\n",inputfile,errormsg);
		}
		dbgprint("read vm_area: 0x%"PRIX64"",vm_area);

		//read first member of vm_area_struct structure
		//this should be mm_struct *, we belong to
		status=img1.read_linear(0, vm_area, &mm_back,
					img1.pointer_size, NULL);
		if ( status ) { 
			errormsg=strerror(errno);
			dbgprint("read() in %s: %s.\n",inputfile,errormsg);
		}
		dbgprint("mm_back: 0x%"PRIX64"",mm_back);
/* 		now we have in mm_struct pointer to vm_area_struct
 * 		should look like this:
 *
 *struct vm_area_struct {
 *  struct mm_struct * vm_mm;       // The address space we belong to. 
 *  unsigned long vm_vm_area;         // Our vm_area address within vm_mm.
 *  unsigned long vm_mm_back;           // The first byte after our mm_back
				    // address within vm_mm. 
 * // linked list of VM areas per task, sorted by address 
 * struct vm_area_struct *vm_next;
 *
 *		vm_area should point back to mm_struct
 * 		So wee need to find linked list (in generic way)
 */
		dbgprint("0x%"PRIX64" vs 0x%"PRIX64"", mm_struct, mm_back);
		if (mm_struct == mm_back) {
			int size=0;
			dbgprint("possible vm_area_struct at 0x%"PRIX64"",vm_area);
			size=simple_list_size(vm_area,
				img1.pointer_size + (img1.integer_size*2));
			if ( size > max ) {
				//save local maximum
				printf("- lin_find_process_VMA: "\
					"vm_area_struct at 0x%"PRIX64", "\
					"size: %d\n", vm_area, size);
				max=size;
				img1.ts_hl_mm_dst=i;
				img1.vm_size_dst=img1.pointer_size; 
				img1.vm_list_dst=img1.pointer_size +
						(img1.integer_size*2);
			}

		}
	}
	dbgprint("mm_struct at %"PRIX64", ts_hl_mm_dst determined as: %d",
		mm_struct, img1.ts_hl_mm_dst);
	debugoff;
	return 0;
}

/* 
 * Function for printing VMA list of process.
 * 
 * list_process_VMA(listhead)
 * listhead is address of list_head of one of the processes
 * 
 * This function should work for any Linux,
 * once determination is completed. 
 *
 * Only printf VMA addresses for now.
 */

int list_process_VMA(u_int64_t addr,FILE * outdump){
	
	u_int64_t vm_area=0,vm_struct=0,start=0,end=0;
	int size=0;
	void *buffer=NULL;


	dbgprint("reading from 0x%"PRIX64", offset: %d",
		addr+img1.ts_hl_mm_dst, img1.ts_hl_mm_dst);

	if (outdump != NULL) {
		 //we will dump memory, so prepare buffer
		if (!(buffer = malloc(MB_1))) { 
			//1MB buffer
			dbgprint("No memory");
			return -2;
		}	
	}	
	
	//first read pointer to mm_struct
	status=img1.read_linear(0, addr+img1.ts_hl_mm_dst, &vm_struct,
				 img1.pointer_size, NULL);
	if ( status ) { 
		errormsg=strerror(errno);
		dbgprint("read() in %s: %s.\n",inputfile,errormsg);
	}	

	dbgprint("read 1: 0x%"PRIX64"",vm_struct);

	//read first member of mm_struct structure
	//this should be vm_area_struct *
	status=img1.read_linear(0, vm_struct, &vm_area, 
				img1.pointer_size, NULL);

	if ( status ) { 
		errormsg=strerror(errno);
		dbgprint("read() in %s: %s.\n",inputfile,errormsg);
	}
	dbgprint("read 2: 0x%"PRIX64"",vm_area);

	for (;;size++){

		dbgprint("ptr: 0x%"PRIX64", size %d",vm_area,size);

		// get start of area 
		status=img1.read_linear(0, vm_area+img1.vm_size_dst,
					&start, img1.integer_size, NULL);
		if ( status ) { 
			errormsg=strerror(errno);
			dbgprint("read() in %s: %s.\n",inputfile,errormsg);
		}

		// get end of area 
		status=img1.read_linear(0, 
				vm_area+img1.vm_size_dst+img1.integer_size,
				&end,
				img1.integer_size,
				NULL);

		if ( status ) { 
			errormsg=strerror(errno);
			dbgprint("read() in %s: %s.\n",inputfile,errormsg);
		}
		
		if ((start == end) && (start == 0)) {
			printf("-- none (kernelspace process?)\n");
			return 0;
		} 
	
		// Terrifying 0x%0*"PRIX64" is for nice padding ;)
		printf("-- 0x%0*"PRIX64"-0x%0*"PRIX64", (",
				img1.pointer_size*2, start,
				img1.pointer_size*2, end); 

		if (outdump != NULL) { 
			// if dumping process memory
			// Dump all areas into one file,
			// ignore distances between areas.
			// TODO coredump format?

			u_int64_t mm_struct=0,msize=0;
			u_int64_t pgd_virt=0,pgd_phys=0;

			// get PGD address first:
			// mm_struct
			status=img1.read_linear(0, 
						addr+img1.ts_hl_mm_dst,
						&mm_struct, 
						img1.pointer_size,
						NULL);
			if ( status ) {

				errormsg=strerror(errno);
				dbgprint("read() in %s: %s.\n",
					inputfile, errormsg);
				return -7; // cannot dump memory
			}

			//virtual address of pgd
			status=img1.read_linear(0,
						mm_struct+img1.mm_pgd_dst,
						&pgd_virt,
						img1.pointer_size,
						NULL);

			if (status) {
				errormsg=strerror(errno);
				dbgprint("read() in %s: %s.\n",
					inputfile, errormsg);
				return -8; // cannot dump memory
			}

			// physical address of pgd
			status=img1.read_linear(0, pgd_virt, NULL,
						0, &pgd_phys);
			if ( status ) {
				errormsg=strerror(errno);
				dbgprint("read() in %s: %s.\n",
					inputfile, errormsg);
				return -9; // cannot dump memory
			}

			// So now we have physical pgd address
			dbgprint("pgd at: %"PRIx64"",pgd_phys);

			msize=end-start;
			dbgprint("Area size: %"PRIu64"",msize);
			
			u_int64_t deb_ph_addr;
			// We need for, because area can be bigger than buffer
			for (;msize>0;) {
				// read content of memory
				status=img1.read_linear(pgd_phys, 
							start,
							buffer,
							(msize < MB_1)?msize:MB_1,
							&deb_ph_addr);
				if ( status ) {
					errormsg=strerror(errno);
					dbgprint("read() in %s: %s.\n",
						inputfile,errormsg);
				}
				dbgprint("Read from ph addr 0x%"PRIX64"",
					deb_ph_addr);
				dbgprint("writing %"PRIu64" bytes to output",
						(msize < MB_1)?msize:MB_1);

				// Write content to file
				fwrite(buffer, (msize < MB_1)?msize:MB_1,
					 1, outdump);

				if (msize < MB_1) { 
					// unsigned cannot be < 0 ;)
					msize=0;
				} else {
					msize=msize - MB_1; 
				}

			}
		}
		
		if (start > end) { 
			// Negative size, something is wrong..
			printf("!!! Negative size, "\
				"something is wrong !!! )\n");
		} else	{	
			// print in human readable format
			human_print(end-start);
			printf(")\n");
		}

	
		if (size > MAX_LIST_SIZE ) {
			printf("List is too long. Something is wrong,"\
				" breaking.");
			return -2; //List too long
		}
		
		// end of list
		status=img1.read_linear(0, vm_area+img1.vm_list_dst,
				&vm_struct, img1.pointer_size, NULL);

		if ( status ) {
			errormsg=strerror(errno);
			dbgprint("read() in %s: %s.\n",inputfile,errormsg);
			return -1; //List should be ended by NULL pointer.
		}
		dbgprint("next list member: %"PRIX64"",vm_struct);	
		if ((vm_area == vm_struct ) || (vm_struct == 0 )) {
			return (size+1); 
			// It is strange, but sometimes list 
			// seems to be finished by pointer to itself.
		}
		vm_area=vm_struct;
	}
	return -5; //Something is wrong
}


/*
 * Function that should list all open files of 
 * 
 * lin_list_files(listhead)
 * return 0 on success 
 * 
 * TODO not working
 * Dirty & Dangerous
 */

int lin_list_files(u_int64_t listhead){
	
	return 0; 
}

/* 
 * Function for creating structure pattern.
 *
 * Just internal help function.
 *
 * first structure is saved "as is",
 * other structures are compared and mask is created. 
 *
 */

int create_struct_pattern(int *first_pattern, 
			u_int64_t *list1,
			unsigned char *pattern,
			unsigned char *pattern_tmp,
			unsigned char *pattern_pos,
			int distance) 
{
	dbgprint("");
	if (*first_pattern == 1) { 
		// save first structure "as is"
		// and diff others structures
		img1.read_linear(0,
				*list1,
				pattern,
				distance,
				NULL);
		*first_pattern=0;
		return 0;
	}

	img1.read_linear(0, *list1, pattern_tmp, distance, NULL);
	int i;
/*debug	   printf("\n");
	   for (i=0;i<distance;i++){
	   printf("%2x",pattern[i]);
	   }
	   printf("\n----------\n");
	   for (i=0;i<distance;i++){
	   if (pattern_pos[i] == 0) {
	   printf("%2x",pattern[i]);
	   } else {
	   printf("..");
	   }
	   }
	   printf("\n");
*/	 

	for (i=0;i<distance;i++) {
		// Test for non-zero values
		if ((pattern_pos[i] == P_HEX) && 
			(pattern[i] != pattern_tmp[i]) &&
			(pattern[i]!=0) && 
			(pattern_tmp[i]!=0) ) {

			//  non-zero value
			pattern_pos[i]=P_NZ; 
		}	

		// Keep only same things
		if ((pattern_pos[i] == P_HEX) && 
			(pattern[i] != pattern_tmp[i])) {

			//  not same for all proc.
			pattern_pos[i]=P_UNK; 
		}
		// No non-zero value
		if ((pattern_pos[i] == P_NZ) &&
			((pattern[i]==0) || 
			 (pattern_tmp[i]==0)) ) {

			// non-non-zero;)
			pattern_pos[i]=P_UNK; 
		}	

	}
	return 0;
}

/* 
 * Function for printing structure pattern.
 *
 * Just internal help function.
 *
 * Print in console friendly format
 *
 */

int print_struct_pattern(int distance,
			char *name,
			unsigned char * pattern,
			unsigned  char * pattern_pos) 
{
	int i;
	printf("----- %s PATTERN START -----\n",name);

	for (i=0;i<distance;i++){

		// I decided leave patterns in one line,
		// so they can be easily cut&pasted
		//	if ( j > LINE_SIZE) { // To cut long lines
		//		j=0;
		//		printf("\n");
		//	}
		switch(pattern_pos[i]){
			case P_HEX: 
				// hex value
				printf("%02x",pattern[i]);
				break;
			case P_UNK:
				// "bgrep" format
				printf("??"); 
				break;
			case P_NZ: 
				// non-zero value
				printf("NZ");
				break;
		}
	}
	printf("\n----- %s PATTERN END -----\n",name);
	return 0;
}	

/* 
 * Function for printing process list.
 * 
 * list_processes(listhead,distance)
 * listhead is address of list_head of one of the processes
 * distance is in bytes from given listhead structure to comm field
 * 
 * This function should work for any OS with processes in list_head,
 * once determination is completed. 
 *
 * Only printf processes for now.
 */

int list_processes(u_int64_t listhead, int distance){
	
	u_int64_t list1=0,list2=0, phys_addr=0;
	u_int64_t counter=0,loop_detection=0,loop_detection_heuristic=0;
	char name[TS_COMM_LEN];
	int loop=1,first_pattern=1;
	int list_areas=0;	// process memory areas

	// for creating pattern
	unsigned char *pattern=NULL,*pattern_pos=NULL,*pattern_tmp=NULL; 

	dbgprint("os: %d",img1.os);
	printf("Listing processes: \n");


	// HOOK
	dbgprint("before hook");
	// Call OS specific hook
	if (img1.os1->hook_list_processes(&list_areas)){
		// something is wrong
		return -2;
	} 
	dbgprint("after hook");

	// Creating process pattern 
	if (do_create_proc_pattern == 1) {
		dbgprint("Creating process pattern");

		//pattern itself
		if ((pattern=(unsigned char *)malloc(distance)) == NULL) {
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		//pattern to compare
		if ((pattern_tmp=(unsigned char *)malloc(distance)) == NULL) { 
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		//list of same bytes
		if ((pattern_pos=(unsigned char *)malloc(distance)) == NULL) {
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		// all bytes are ok at beginning
		memset(pattern_pos,P_HEX,distance); 
	}



	list1 = listhead;
	
	while (loop) { 
		// This is way "down" in list_head (or "up" if bsd-like)
		img1.read_linear(0, 
				list1,
				&list2, 
				img1.pointer_size,
				&phys_addr);
		dbgprint("read 0x%"PRIX64" from 0x%"PRIX64"",list2,list1);

		if ((list1 == list2) || (list2 == 0)) {
			// we point back to ourselves => circle is closed
			// This is fine for bsd
			break;
		}

		if (loop_detection_heuristic == 0) {	
			// If we process same task_struct twice.
			//  Result not guaranteed.
			loop_detection_heuristic = list2;
		} else {
			if (loop_detection_heuristic == list2) {
				break;
			}
		}

		list1 = list2;
		counter++;	//number of processes found
		if (loop_detection++ > MAX_PROC_FOUND) { 
			// there cannot be more than 65K 
			// processes on Linux/Windows (ehm;)
			printf("- list_processes: !!! "\
				"Infinite loop detected !!! "\
				"Something is wrong\n");
			if (do_create_proc_pattern == 1) {
				free(pattern);
				free(pattern_pos);
				free(pattern_tmp);
			}
			return -3;
		};

		dbgprint("list1: 0x%"PRIX64", list1+distance: 0x%"PRIX64"",
			list1, list1+distance);	
		
		// read next entry
		img1.read_linear(0,
				list1+distance,
				name,
				TS_COMM_LEN,
				NULL);
		printf("- proc addr (v/p: 0x%"PRIX64"/0x%"PRIX64") "\
				"name: %s", list1, phys_addr, name); 

		if (do_create_proc_pattern == 1) {
			create_struct_pattern(&first_pattern,
					&list1,
					pattern,
					pattern_tmp,
					pattern_pos,
					distance);
		}

		// Default is yes, but user can explicitly disable this
		if (list_areas==1) { 
			printf("\n-- mapped areas:\n");
			list_process_VMA(list1,NULL);
		}

		// Now, (if requested) we will print list of all 
		// file descriptors for this process
		if ( do_files_list > 0 ) {
			// This branch is not working right now.  TODO
			//printf(", open files:\n");
			printf("\n");
			//lin_list_files(listhead); 
			//list files for this process
		} else {
			printf("\n");
		}
	}	

	if (do_create_proc_pattern == 1) {
		// print process pattern in function
		print_struct_pattern(
			distance,
			"PROCESS",
			pattern,
			pattern_pos);

		free(pattern);
		free(pattern_pos);
		free(pattern_tmp);
	}

	printf("Listing processes done.\n");

	return counter;
}

/*
 *	We do not search for in init directly, 
 *	because it is too common string in memory => takes too long
 *	This is almost clone of list_processes
 *	
 *	lin_find_userspace_process()
 *	return 0 on success, and set data in global structure
 */

int lin_find_userspace_process(){
	
	u_int64_t list1=0,list2=0,phys_addr=0;
	u_int64_t counter=0,loop_detection=0,loop_detection_heuristic=0;
	char name[TS_COMM_LEN];
	int loop=1,status=0;

	printf("- lin_find_userspace_process: Looking for userspace "\
		"process \"%s\".\n",img1.magic_user_process); 
	
	//This is already from list_head, but not userspace
	list1 = img1.ts_one_of_lists; 

	while (loop) { //This is way "down" in list_head
	
		img1.read_linear(0, list1, &list2, img1.pointer_size, NULL);

		dbgprint("l1:%"PRIX64", l2:%"PRIX64".\n",list1,list2); 

		if ((list1 == list2) || (list2 == 0)) { 
			// List is empty, 
			// or invalid linear address was provided
			printf("- lin_find_userspace_process: strange,"\
					"list1: %"PRIX64",list2: %"PRIX64"\n",
					list1, list2);
			loop=0;
			break;
		}

		// Otherwise, 
		// we point back to ourselves => circle is closed	
		if (loop_detection_heuristic == 0) {	
			// If we process same task_struct twice. 
			// No guaranteed result
			loop_detection_heuristic = list2;
		} else {
			if (loop_detection_heuristic == list2) {
				printf("- lin_find_userspace_process: "\
					"Not found \"%s\", something is "\
					"wrong.\n", img1.magic_user_process);
				return -1;
			}
		}
		list1 = list2;
		counter++;	//number of processes found
		if (loop_detection++ > MAX_PROC_FOUND) { 
			//there cannot be more than 65K processes on Linux
			printf("- lin_find_userspace_process: !!! "\
				"Infinite loop detected !!! "\
				"Something is wrong\n");
			return -3;
		};

		img1.read_linear(0, list1+img1.ts_hl_comm_dst,
					name, TS_COMM_LEN, &phys_addr);
		dbgprint("Userspace process \"%s\" at 0x%"PRIX64"\n",
			name, list1);
		dbgprint("%s",img1.magic_user_process);

		status=strncmp(name, img1.magic_user_process, TS_COMM_LEN);
		if ( status == 0 ) {
			printf("Found userspace process \"%s\" at "\
					"(v/p) 0x%"PRIX64"/0x%"PRIX64"\n",
					img1.magic_user_process, 
					list1,
					phys_addr); 

			img1.ts_one_of_lists=list1;
			return 0;
		}
	}

	return -5; //Something is wrong, list is empty or invalid
}

/*  
 *  Function for getting size of circular list head .
 *  ( used in linux/win process structures )
 *
 *  Whole list is traversed, until first member is seen 
 *  second time. Or until loop detection is successful.
 * 
 *  input: Physical address of list_head
 *  output: Number of members, negative value if problem is detected
 */

int lin_list_head_size(u_int64_t listhead){
	
	u_int64_t list1=0,list2=0;
	u_int64_t back_list1=0,back_list2=0; // backward checking
	u_int64_t counter=0,errcounter=0; // entries & errors
	u_int64_t loop_detection=0,loop_detection_heuristic=0; 
	int loop=1;

	if ( (status=lseek(inputfd,listhead,SEEK_SET)) == -1 ) {  
		errormsg = strerror(errno);
		fprintf(stderr, "lin_list_head_size: lseek() to 0x%"PRIX64" "\
			"failed because of %s.\n", listhead, errormsg);
		return(-1);
	}
	
	if ( (status=read(inputfd,&list1,img1.pointer_size)) == -1 ) {
		errormsg=strerror(errno);
		fprintf(stderr, "lin_list_head_size: read() in %s: %s.\n",
			inputfile, errormsg);
		return(-2);
	}
	dbgprint(" 0x%"PRIX64"\n",listhead);

	while (loop) { 
		// This is way "down" in list_head

		// forward pointer
		img1.read_linear(0, list1, &list2, img1.pointer_size,NULL);
		// backward pointer
		img1.read_linear(0, list1+img1.pointer_size, &back_list1,
					img1.pointer_size, NULL); 

		// if we have already seen this process (circle closed)
		if ((list1 == list2) || (list2 == 0)) {loop=0;} else 
		{
			// If we process same task_struct twice. 
			// No guaranteed result
			if (loop_detection_heuristic == 0) {	

				loop_detection_heuristic = list2;
			} else {
				if (loop_detection_heuristic == list2) { 
					// inner circle closed
					dbgprint("counter:%"PRIu64".",counter);
					return counter;
				}
			}
			// Check if we are really in double linked list,
			// or just jumping  over some addresses
			if ((back_list1 != back_list2) && (back_list2 != 0)) {
				// tolerate some number of errors
				errcounter++;
				dbgprint("errcounter ++ (%"PRIu64"), bl1: %"PRIX64","\
					"bl2: %"PRIX64", l2: %"PRIX64",l1: %"PRIX64"",
					errcounter, back_list1, back_list2, 
					list2, list1);	

				if (errcounter > img1.os1->ts_magic_list_max_err){
					dbgprint("errcounter reached (%"PRIu64")",
						errcounter);	
					return -7;
				}
			}

			back_list2 = list1; //back pointer checking
			list1 = list2;
			counter++;	//number of processes found
			if (loop_detection++ > MAX_PROC_FOUND) { 
				// there cannot be more than 65K processes 
				// on Linux (ehm;)
				printf("- lin_list_head_size: !!! "\
					"Infinite loop detected !!! "\
					"Something is wrong\n");
				return -3;
			};
	
		}	
	}

	return -5; //Something is wrong
}

/*  
 *  Function for getting size of bsd-like list.
 *  ( used in BSD "proc" structure )
 *
 *                   ||
 *                   ||
 *                   \/ 
 *  INIT |--------> NEXT |-------->  |-------->  NULL
 *  NULL <--------| PREV <--------|              ...
 *
 *
 * from sys/queue.h
 * ...
 * #define LIST_HEAD(name, type)
 * struct name {
 *         struct type *lh_first;  // first element
 * }
 * 
 * #define LIST_HEAD_INITIALIZER(head)
 *         { NULL }
 * 
 * #define LIST_ENTRY(type)
 * struct {
 *         struct type *le_next;   // next element
 *         struct type **le_prev;  // address of previous next element
 * }
 * ...
 *
 *  bsd_list_head_size (address to verify , 
 * 		output parameter "first member of list" )
 *  return number of members on success, otherwise error < 0
 *  if list_head_start is not NULL, pointer to first member of 
 *  list is stored here.
 * 
 *
 *  Because this list is not cycled, first we must travel to one side
 *  until NULL is encountered, and then travel to the other side,
 *  looking for NULL pointer. We also need to propagate information 
 *  where list starts
 *
 *  input: Physical address of list_head, offset to string (optional)
 *  output: Number of members ( <0 if error occurs ), start of list (optional), 
 *          total size of all strings at offset (optional)
 * 
 */

int bsd_list_head_size(u_int64_t listhead, 
			u_int64_t * list_head_start, 
			int offset, 
			int * strings_size){
	
	u_int64_t list1=0,list2=0;
	u_int64_t back_list1=0,back_list2=0; //backward checking
	u_int64_t phys_list1=0,phys_list2=0;
	u_int64_t loop_detection_heuristic=0;
	int counter=0,errcounter=0;
	int loop=1;
	char name[TS_COMM_LEN]; //structure name (string)

	dbgprint("listhead: 0x%"PRIX64", offset %d, (list_head_start: 0x%"PRIX64", "\
		 "strings_size: %d) ", 
		listhead,offset, *list_head_start, *strings_size);

	if ( (status=lseek(inputfd,listhead,SEEK_SET)) == -1 ) {  
		errormsg = strerror(errno);
		fprintf(stderr, "lin_list_head_size: lseek() to 0x%"PRIX64" "\
			"failed because of %s.\n", listhead, errormsg);
		return(-1);
	}
	
	if ( (status=read(inputfd,&list1,img1.pointer_size)) == -1 ) {
		errormsg=strerror(errno);
		fprintf(stderr, "lin_list_head_size: read() in %s: %s.\n",
			inputfile, errormsg);
		return(-2);
	}
	dbgprint(" 0x%"PRIX64", list1: 0x%"PRIX64"",listhead,list1);

	// counting length of all list_members strings. 
	// See explanation before this function.
	if ((offset != 0) && (strings_size != NULL)) {
		*strings_size=0;
	}

	loop_detection_heuristic = list1; //loop detection
	// first up and first down back-pointer checks will fail
	// ok, just ignore them


	while (loop) { //This is way "up" in list_head

		// forward pointer
		img1.read_linear(0, list1, &list2, 
				img1.pointer_size, &phys_list1);
		// backward pointer
		img1.read_linear(0, list1+img1.pointer_size,
				&back_list1, img1.pointer_size, &phys_list2);
		dbgprint("up-step, 0x%"PRIX64" (0x%"PRIX64"), 0x%"PRIX64" (0x%"PRIX64")",
			 list2, phys_list1, back_list1, phys_list2);

		// Counting length of all list_members strings. 
		// See explanation before this function.
		if (offset != 0) {
			// entry name
			img1.read_linear(0, list1+offset, name,
						TS_COMM_LEN, &phys_list1);
			*strings_size+=strnlen(name,TS_COMM_LEN);
			dbgprint("List 0x%"PRIX64" at 0x%"PRIx64" (0x%"PRIX64") proc: %s",
				list1, list1+offset, phys_list1, name);
		}

		if ((list1 == list2) || (list2 == 0)) {

			// we have already seen this process or NULL
			loop=0;
			break;
		}

		if (loop_detection_heuristic == list2) { 
			//circle closed
			printf("- bsd_list_head_size: Susp. "\
				"struct at 0x%"PRIX64", loop "\
				"(%d).\n", list1, counter);
			return -9;
		}

		// Check if we are really in double linked list, 
		// or just jumping over some addresses
		if ((back_list1 != back_list2) && (back_list2 != 0)) {
			//tolerate some number of errors
			errcounter++;

			dbgprint("errorcounter ++ (%d), bl1: %"PRIX64",bl2:"\
				" %"PRIX64", l2: %"PRIX64",l1: %"PRIX64"", errcounter,
				 back_list1, back_list2, list2, list1);	

			if (errcounter > img1.os1->ts_magic_list_max_err){
				dbgprint("errorcounter reached (%d)",
					errcounter);
				return -7;
			}
		}

		back_list2 = list1; //back pointer checking
		list1 = list2;
		counter++;	//number of processes found
		if (counter > MAX_PROC_FOUND) { 
			// there cannot be more than 65K processes 
			// on BSD (ehm;)
			printf("- bsd_list_head_size: !!! Infinite "\
				"loop detected !!! Something is wrong\n");
			return -3;
		};
	

	}

	// Do same thing for way "down"
	dbgprint("down: 0x%"PRIX64", list1: 0x%"PRIX64"\n",listhead,list1);
	list1 = loop_detection_heuristic;	//original pointer
	loop=1;

	while (loop) { 
		// This is way "down" in list_head

		// forward pointer
		img1.read_linear(0, list1, &back_list1, 
					img1.pointer_size, &phys_list2);
		// backward pointer
		img1.read_linear(0, list1+img1.pointer_size, &list2, 
					img1.pointer_size, &phys_list1);
		dbgprint("down-step, 0x%"PRIX64" (0x%"PRIX64"), 0x%"PRIX64" (0x%"PRIX64")",
			 list2, phys_list1, back_list1, phys_list2);

		// counting length of all list_members strings. 
		// See explanation before this function.
		if (offset != 0) {
			// entry name 
			img1.read_linear(0, list1+offset, name, 
						TS_COMM_LEN, &phys_list1);
			*strings_size+=strnlen(name, TS_COMM_LEN);
			dbgprint("List 0x%"PRIX64" at 0x%"PRIx64" (0x%"PRIX64") proc: %s",
				list1, list1+offset, phys_list1, name); 
		}

		if ((list1 == list2) || (list2 == 0)) {
			// we have already seen this process or NULL
			loop=0; 
			break;
		}

		if (loop_detection_heuristic == list2) { 
			//circle closed
			return -9;
		}

		// Check if we are really in double linked list, 
		// or just jumping over some addresses
		if (back_list1 != back_list2) {
			//tolerate some number of errors
			errcounter++;
			dbgprint("errorcounter ++ (%d), bl1: %"PRIX64",bl2: "\
				"%"PRIX64", l2: %"PRIX64",l1: %"PRIX64"",errcounter, 
				back_list1, back_list2, list2, list1);	
			if (errcounter > img1.os1->ts_magic_list_max_err){
				dbgprint("errorcounter reached (%d)",
					errcounter);
				return -7;
			}
		}

		back_list2 = list1; //back pointer checking
		list1 = list2;
		counter++;	//number of processes found
		if (counter > MAX_PROC_FOUND) { 
			//there cannot be more than 65K processes on BSD (ehm;)
			printf("- bsd_list_head_size: !!! Infinite "\
				"loop detected !!! Something is wrong\n");
			return -3;
		};
	
	
	}
	
	// return first member of list
	if (list_head_start != NULL) {
		* list_head_start=list1;
	}
	return counter; // whole size
}



/*
 * Function for printing list of modules.
 *
 * lin_list_modules(listhead, distance)
 * listhead is address of list_head of one of the modules 
 *(not linear address, but address in physical dump file)
 * distance is in bytes from given listhead structure 
 * to comm (char[16]) field.
 *
 * Only printf modules for now.
 */


int lin_list_modules(u_int64_t listhead,int distance){
	
	u_int64_t list1=0,list2=0,phys_addr=0;
	u_int64_t counter=0,loop_detection=0,loop_detection_heuristic=0;

	//according to module.h, with  reserve
	char name[MD_NAME_LEN]; 
	int loop=1,first_pattern=1;

	// for creating pattern
	unsigned char *pattern=NULL,*pattern_pos=NULL,*pattern_tmp=NULL;

	// Creating module pattern 
	if (do_create_module_pattern == 1) {
		dbgprint("Creating module pattern");
		// pattern itself
		pattern=(unsigned char *)malloc(
					distance+img1.os1->md_pattern_spread);
		if (pattern == NULL) {
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		// pattern to compare
		pattern_tmp=(unsigned char *)malloc(
					distance+img1.os1->md_pattern_spread);

		if (pattern_tmp == NULL) {
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		// list of same bytes
		pattern_pos=(unsigned char *)malloc(
					distance+img1.os1->md_pattern_spread);
		if (pattern_pos == NULL) {
 			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}

		// all bytes are ok at beginning
		memset(pattern_pos,P_HEX,distance+img1.os1->md_pattern_spread);
	}


	printf("Listing modules:\n");
	list1 = listhead;

	while (loop) { // This is way "down" in list_head
	
		img1.read_linear(0, list1, &list2, 
					img1.pointer_size, NULL);

		if (do_create_module_pattern == 1) {
			// use function for pattern creation
			create_struct_pattern(&first_pattern,
						&list1,
						pattern,
						pattern_tmp,
						pattern_pos,
					distance+img1.os1->md_pattern_spread);
		}



		if ((list1 == list2) || (list2 == 0)) {
			loop=0;
			break;
		}
		
		// we point back to ourselves => circle has closed	

		// If we process same module twice. No guaranteed result
		if (loop_detection_heuristic == 0) {
			loop_detection_heuristic = list2;
		} else {
			if (loop_detection_heuristic == list2) {
				if (do_create_module_pattern == 1) {
					// print pattern in function
					print_struct_pattern(
					distance+img1.os1->md_pattern_spread,
						"MODULE",
						pattern,
						pattern_pos);

					free(pattern);
					free(pattern_pos);
					free(pattern_tmp);
				}

				printf("Listing modules done.\n");
				return counter;
			}
		}
		list1 = list2;
		counter++;	// number of modules found
		if (loop_detection++ > MAX_PROC_FOUND) { 
			// there cannot be more than 65K modules 
			// on Linux (ehm;)
			printf("- list_modules: !!! Infinite loop "\
				"detected !!! Something is wrong\n");
			return -3;
		};

		img1.read_linear(0, list1+distance, name, 
					MD_NAME_LEN, &phys_addr);
		printf("- module addr (v/p): 0x%"PRIX64"/0x%"PRIX64", "\
			"name: %s\n", list1, phys_addr, name); 	
	}

	return -5; // Something is wrong
}

/*
 * Function that tries to determine processes specific information 
 * about analyzed image.
 *
 * Some process (and if analyzing linux then some userland process) 
 * are found, and distance between important fields are determined.
 *
 * If successful enough, OS is determined as side effect.
 *
 * Information about "module" structure are determined separately,
 * because module support is only optional, and on some systems
 * there are almost no modules.
 *
 * proc_determine ()
 * return 0 on success, otherwise error <> 0
 * Discovered information are saved in global structure img1.
 *
 * */

int proc_determine(){
	
	u_int64_t process=0;

	printf("- proc_determine: Looking for process \"%s\".\n",
		img1.magic_process);

	if (search_init(inputfile,img1.magic_process,0) != 0 ) { 
		// This is not optimal
		printf("proc_determine: search_init returned error."\
			" Aborting.\n");
		return -1;
	}

	while ((process=search_next())!=0){
		dbgprint("Trying to verify: 0x%"PRIX64" as task_struct",process);
		if (verify_task_struct(process) == 0) { 
			//we have found valid process
			printf("Found valid process at 0x%"PRIX64".\n",process);
			printf("Determine list-name offset as %d\n",
				img1.ts_hl_comm_dst);
			break;
		}
	}
	dbgprint("before hook1");
	// Call OS specific hook
	if (img1.os1->hook_proc_determine(&process) != 0) {
		// process not found
		return 2;
	} 
	dbgprint("after hook1");

	// Call OS specific hook
	// Process structure is found at this moment, but for some OS
	// additional information needs to be determined 
	// (userspace process, mm_struct, ...)

	if (img1.os1->hook_proc_determine2() != 0) {
		// something is wrong so work in "optimistic" mode and
		// print all we can
	} 
	dbgprint("after hook2");

	dbgprint("img1.os: %d",img1.os);

	printf("Determination process of task_struct complete.\n");
	return 0;
}


/*
 * Function that tries to determine linux module-specific 
 * information about analyzed image.
 *
 * Some module is found,
 * and distance between important fields are determined.
 *
 * This is done separately,
 * because module support is only optional, and is not 
 * critical for another analyze.
 *
 * linux_determine_module ()
 * return 0 on success, otherwise error <> 0
 * Discovered information are saved in global structure img1.
 *
 * */

int lin_linux_determine_module(){
	
	u_int64_t addr=0;
	int status;
	
	printf("- linux_determine_module: Looking for module \"%s\".\n",
		img1.magic_module);
	printf("- linux_determine_module: Starting at address 0x%X.\n",
		MODULE_MAGIC_OPT_ADDR);

	status = search_init(inputfile, 
				img1.magic_module,
				MODULE_MAGIC_OPT_ADDR);

 	if ( status ) { 
		printf("linux_determine: search_init returned error."\
			" Aborting.\n");
		return -1;
	}

	
	while ((addr=search_next())!=0){
		dbgprint("Trying to verify: %"PRIX64" as task_struct\n",addr);
		if (lin_verify_module(addr) == 0) { 

			//we have found valid module
			printf("Found valid module at 0x%"PRIX64".\n",addr);
			printf("Determination process of module "\
				"structure complete.\n");
			return 0;
		}
	}

	//Found nothing
	printf("- linux_determine_module: Cannot find module %s"\
		" while using optimization.\n",img1.magic_module); 
	printf("- linux_determine_module: Retrying with whole "\
		"file (slower).\n");

	if (search_init(inputfile,img1.magic_module,0) != 0 ) { 
		//Search whole file
		printf("linux_determine: search_init returned error."\
			" Aborting.\n");
		return -1;
	}
	
	while ((addr=search_next())!=0){
		if (lin_verify_module(addr) == 0) { 
			//we have found valid module
			printf("Found valid module at 0x%"PRIX64".\n",addr);
			printf("Determination process of module structure "\
				"complete.\n");
			return 0;
		}
	}

	//Found nothing
	printf("- linux_determine_module: Cannot find module %s.\n",
		img1.magic_module); 
	return 2;
}

/*
 * Function that "safely" prints string
 * ie. dots instead of strange characters
 *
 * Just internal helper function.
 *
 * safeprint (name, length)
 * return 0 on success, otherwise error <> 0
 */

int safeprint(unsigned char *name, int len) {
	int i=0;
	for(;i<=len;i++) {
		int j=name[i];
		if (isprint(j) && ( j == ' ' || !isspace(j)) ) {
			// only printable chars, 
			// but not end of lines, 
			// but yes space;)
			putc(j,stdout);
		} else {
			//non printable (in ascii)
			putc('.',stdout); 
		}
	}
	return 0;
}

/*
 * Function that prints detailed information 
 * about certain process. 
 *
 * If phys_process is not zero, physical address is used,
 * (and virtual is ignored)
 * otherwise virtual address is used.
 * 
 * print_process_info ( process address, 
			process physical address,
			filename to store process memory)
 * return 0 on success, otherwise error <> 0
 *
 */

int print_process_info(u_int64_t virt_process, 
			u_int64_t phys_process, 
			char * dump_filename)
{
	FILE *outdump;
	int status=0,i=0;	
	u_int64_t phys_addr=0,mmstruct=0,pgd_lin=0,pgd_phys=0;

	//pointer to previous and next process
	u_int64_t prev_lin=0,prev_phys=0,next_lin=0,next_phys=0; 
	unsigned char name[TS_COMM_LEN];

	
	if ( phys_process != 0) {
		// process is specified by physical address

		// we need to get process virtual address
		// read pointer to next list_head	
		if ( (status=lseek(inputfd,phys_process,SEEK_SET)) == -1 ) {
			errormsg = strerror(errno);
			dbgprint("lseek() to 0x%"PRIX64" "\
				"failed because of %s.\n", 
				phys_process, errormsg);
			return(1);
		}

		status=read(inputfd, &next_lin, img1.pointer_size);
		if ( status == -1 ) {
			errormsg=strerror(errno);
			dbgprint("read() in %s: %s.\n",
				inputfile, errormsg);
			return(2);
		}

		// virtual address of ourselves
		status=img1.read_linear(0, 
					next_lin + img1.pointer_size, 
					&virt_process, 
					img1.pointer_size, 
					NULL);
		// so we will print it correctly for both cases
		phys_addr = phys_process;

	} else {
		// process is specified by virtual address

		// physical address of list_head
		status=img1.read_linear(0, virt_process, 
					NULL, 0, &phys_addr);
	}
	// process name
	status=img1.read_linear(0, virt_process+img1.ts_hl_comm_dst,
					name, TS_COMM_LEN, NULL);
	//mm_struct
	status=img1.read_linear(0, virt_process+img1.ts_hl_mm_dst,
					&mmstruct, img1.pointer_size ,NULL);

	dbgprint("mmstruct at %"PRIX64", ts_hl_mm_dst: %d, mm_pgd_dst: %d",
		mmstruct, img1.ts_hl_mm_dst, img1.mm_pgd_dst);
	if (status == 0 ) {
		//pgd
		img1.read_linear(0, mmstruct+img1.mm_pgd_dst, &pgd_lin, 
					img1.pointer_size, &pgd_phys);
		//physical address of pgd
		status=img1.read_linear(0, pgd_lin ,NULL,0, &pgd_phys);
	} 

	//ptr to previous proc
	img1.read_linear(0, virt_process, &prev_lin, 
				img1.pointer_size, &prev_phys); 
	//ptr to next proc
	img1.read_linear(0, virt_process+img1.pointer_size, &next_lin,
				img1.pointer_size, &next_phys);

	printf("Detailed information about process at 0x%"PRIX64".\n",
		virt_process);
	printf("Process linear address(in kernel context): 0x%"PRIX64"\n",
		virt_process); 
	printf("Process physical address(offset in dump): 0x%"PRIX64"\n",
		phys_addr);
	printf("Process name(short): \"%s\"\n",name);
	printf("Process PGD address(linear): 0x%"PRIX64"\n",pgd_lin);
	printf("Process PGD address(physical): 0x%"PRIX64"\n",pgd_phys);
	printf("Previous process: 0x%"PRIX64" (phys: 0x%"PRIX64")\n",prev_lin,prev_phys);
	printf("Next process: 0x%"PRIX64" (phys: 0x%"PRIX64")\n",next_lin,next_phys);
	printf("Process mapped areas (userspace addresses):\n");
	if (dump_filename == NULL) {
		list_process_VMA(virt_process,NULL);
	} else {	
		printf("Saving process memory into %s:\n",dump_filename);

		if ( (outdump = fopen(dump_filename,"w")) == NULL ) {
			errormsg = strerror(errno);
			fprintf(stderr, "%s: %s.\n",statusfile,errormsg);
			return 22;
		}

		list_process_VMA(virt_process,outdump);

		i=fclose(outdump);

		printf("Process memory saved to file: %s.\n",dump_filename);

	}
	printf("Process open files: TODO;)\n");
//      i386_lin_list_files(listhead); //list files for this process //TODO
	return i;

}

/*
 * Function that prints detailed information 
 * about certain module.
 *
 * print_module_info ()
 * return 0 on success, otherwise error <> 0
 *
 * */

int print_module_info(u_int64_t single_module)
{
	u_int64_t phys_addr=0;		
	unsigned char name[MD_NAME_LEN];

	img1.read_linear(0, single_module+img1.md_hl_name_dst,
				name, MD_NAME_LEN, &phys_addr); 
	printf("Detailed information about module at 0x%"PRIX64".\n",
		single_module);
	printf("Module linear address(in kernel context): 0x%"PRIX64"\n",
		single_module); 
	printf("Module physical address(offset in dump): 0x%"PRIX64"\n",
		phys_addr);
	printf("Module name(short): \"");

	safeprint(name,MD_NAME_LEN);
	printf("\"\n");	
	return 0;	
}
