#ifndef _ALLARCH_FUNCTIONS_H
#define _ALLARCH_FUNCTIONS_H

/*
 *	Functions for generic image analysing 
*/

#include "../globals.h"
#include "i386/functions_i386.h"
#include "x86_64/functions_x86_64.h"
#include "arm/functions_arm.h"

/*
 * Function that sets architecture
 * settings according to values of supplied structure 
 *
 * arch_settings ()
 * return 0 on success, otherwise error > 0
 * results are stored in global structure
 *
 */

int arch_settings();

/*
 * Function that sets pointer to correct functions,
 * according to detected architecture
 *
 */


int set_arch_from_list(char *arch_name);


/*
 * Function with heuristic for architecture guessing
 *
 * guess_architecture ()
 * return 0 on success, otherwise error > 0
 * results are stored in global structure
 *
 */

int guess_architecture();


/*
 * Function for verifying if given address contains structure list_head.
 * (cyclic version)
 *
 * Verification is done by finding and checking double linked
 * "list_head". This way is chosen instead of classic
 * pattern matching, because I believe it is more flexible and
 * can be used for almost every 2.6 kernel version. (even future;)
 * It does not depend on task_struct definition too much, only
 * important characteristic is number of list_heads between
 * "comm" array and list_head "tasks". This is specified by
 * TS_MAGIC_HEADLIST_NUMBER.
 *
 * Correct list_head can be verified easily:
 * Data at address point to (virtual) address where is stored another
 * list_head. Second pointer in this second head list point back
 * to (virtual) address where we started.
 *
 *  ...   |--------> |---------->  |---------->   ...
 *          <------  | <-----------| <--------
 *
 *
 *
 *  verify_cyclic_list_head (address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */

int verify_cyclic_list_head(u_int64_t ver_addr);


/*
 *  verify_vfsmount (pointer to vfsmount structure)
 *  return 0 on success, otherwise -1
 *
 *  vfsmount structure is "recognized" by simple test:
 *  should contain pointer to string "/dev/".
 *  (char *mnt_name)
 *  This is alone not a strong heuristic, but in combination
 *  with specific chain where structure should be found when trying 
 *  to list open files, it is enough.
 */

int lin_verify_vfsmount(u_int64_t ver_addr);

/*
 *  verify_qstr (pointer to "qstr" structure )
 *  return 0 on success, otherwise error
 *
 *  Used for listing of open files per process.
 *    d_name -> name
 *    (qstr)   (char)
 *
 *  This function verifies, if pointer points to qstr structure.
 *  qstr is elemental structure, and we assume, that will not change (there is no reason)
 *  Verification is done by checking, if string that "name" points to is really "len"
 *  bytes long (looking for \0) and "len" is not 0.
 *
 *  This is qstr definition:
 *
 *  struct qstr {
 *          unsigned int hash;
 *          unsigned int len;
 *          const unsigned char *name;
 *  };
 *                          
 *
 */

int lin_verify_qstr(u_int64_t ver_addr);
/*
 *  verify_dentry (pointer to "dentry" structure )
 *  return 0 on success, otherwise -1
 *  determined distances are saved into image1
 *
 *  Used for listing of open files per process.
 *  Chain we are following looks like:
 *
 *  => task_struct ->       files   ->      fdt     ->      fd      ->      f_path  ->      dentry  ->      d_name-> name
 *  		    (files_struct)      (fdtable)        (file **)          (path)          (dentry)        (qstr)  (char)
 *
 *  This function determines distances in structure dentry
 *  It is better to determine structures by more "sub" heuristics, because amount of tries needed is decreased.
 *  (and amount of false positives too)
 *  
 */

int lin_verify_dentry(u_int64_t ver_addr); 

/*
 *  determine_up_to_path (pointer to "tasks" list_head in task_struct )
 *  return 0 on success, otherwise -1
 *  determined distances are saved into image1
 *
 *  function that tries to determine distances in structures needed
 *  for listing of open files per process up to f_path structure.
 *  Chain we are following looks like:
 *
 *  => task_struct ->       files   ->      fdt     ->      fd      ->      f_path  ->      dentry  ->      d_name-> name
 *  		    (files_struct)      (fdtable)        (file **)          (path)          (dentry)        (qstr)  (char)
 *
 *  So this function determines distances in structures files_struct, fdtable, file and path.
 *  It is better to determine structures by more "sub" heuristics, because amount of tries needed is decreased.
 *  (and amount of false positives too)
 *  
 */

int lin_determine_up_to_path(u_int64_t ver_addr);

/*
 *  determine_open_files (pointer to "tasks" list_head in task_struct )
 *  return 0 on success, otherwise -1
 *
 *  function that tries to determine distances in all structures needed
 *  for listing of open files per process. 
 *  This will definitely not work in 2.4 kernels.
 */

int lin_determine_open_files(u_int64_t ver_addr);

/*
 *  verify_task_struct (pointer to "comm" field of task_struct )
 *  return 0 on success, otherwise -1
 */

int verify_task_struct(u_int64_t ver_addr);

/*
 *
 *  Interesting fields in "module" are just near. This should not change (no reason),
 *  so this function is much simpler then verify_task_struct .
 *
 *  verify_module (pointer to "name" field of module )
 *  return distance between name and list_head in bytes on success, otherwise < 0
 *
 *  
 */

int lin_verify_module(u_int64_t ver_addr);


/* Only printf processes for now.
 * listhead is address of list_head of one of the processes
 * distance is in bytes from given listhead structure to comm (char[16]) field
 */

int list_processes(u_int64_t listhead,int distance);

/*
 * We do not search for in init directly, because it is too common string in memory => takes too long
 * This is almost clone of list_processes, but linux specific.
 * In windows, all processes are already in first list.
 */

int lin_find_userspace_process();

/* Another clone of list_processes
 * Only number of members of list_head
 *
 * input: Physical address of list_head
 * output: Number of members, -3 if infinite loop detected
 */

int lin_list_head_size(u_int64_t listhead);

/* Another clone of list_processes (BSD version)
 * Only number of members of list_head
 *
 * input: Physical address of list_head, offset to string (optional)
 * output: Number of members ( <0 if error occurs ), start of list (optional), 
 *         total size of all strings at offset (optional)
 */

int bsd_list_head_size(u_int64_t listhead, u_int64_t * list_head_start, int offset, int * strings_size);

/* Only printf modules for now.
 * listhead is address of list_head of one of the modules (not linear address, but address in physical dump file)
 * distance is in bytes from given listhead structure to comm (char[16]) field
 */


int lin_list_modules(u_int64_t listhead,int distance);

/*
 * Function that tries to determine linux-specific information 
 * about analyzed image.
 *
 * Some process and then some userland process are found,
 * and distance between important fields are determined.
 *
 * Information about "module" structure are determined separately,
 * because module support is only optional.
 *
 * linux_determine ()
 * return 0 on success, otherwise error <> 0
 * Discovered information are saved in global structure image1.
 *
 * */

int proc_determine();

/*
 * Function that tries to determine processes specific information 
 * about analyzed image.
 *
 * Some process (and if linux then some userland process) are found,
 * and distance between important fields are determined.
 *
 * If sucessfull enough, OS is determined as side effect.
 *
 * Information about "module" structure are determined separately,
 * because module support is only optional.
 *
 * proc_determine ()
 * return 0 on success, otherwise error <> 0
 * Discovered information are saved in global structure image1.
 *
 * */

int lin_linux_determine_module();

/*
 * Function that prints detailed information 
 * about certain process.
 *
 * print_process_info ( process address, process physicall address, 
			filename to store process memory)
 * return 0 on success, otherwise error <> 0
 *
 * */

int print_process_info (u_int64_t virt_process,
			u_int64_t phys_process,
			 char * dump_filename);

/*
 * Function that prints detailed information 
 * about certain module.
 *
 * print_module_info ()
 * return 0 on success, otherwise error <> 0
 *
 * */

int print_module_info(u_int64_t single_module);

/*
 * Function for finding VMAs of given process 
 *
 * lin_find_process_VMA (linear address of "tasks" list_head)
 * return 0 on success, otherwise error > 0
 * Discovered distances are saved in global structure image1.
 *
 * */

int lin_find_process_VMA(u_int64_t my_addr);

/* 
 * Function for finding VPT of given process 
 * (resp. PaGeDirectory).
 *
 * NOTE: ts_hl_mm_dst distance must be known before calling this function.
 * 
 */

int lin_find_process_VPT(u_int64_t my_addr);


#endif
