// Written in April 2009 by Felix Domke <tmbinc@elitedvb.net>
//
// Placed in the public domain April 2009 by the author: no copyright is
// claimed, and you may use it for any purpose you like.
//
// No warranty for any purpose is expressed or implied by the author. 
// Report bugs and send enhancements to the author.
//
// Expanded in 2010 by niekt0@hysteria.sk
// Modified version under GNU GPL

// TODO
// naive search implementation is used
// Rewrite using better algorithm? 
// We need fuzzy search, but only with substitutions (no delete, no append)
// (bitap have problem with large patterns and big alphabet, 
// does even exist better algorithm?)


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <getopt.h>

#define pattern_size 0x1000

int ascii2hex(char c)
{
	if (c < '0')
		return -1;
	else if (c <= '9')
		return c - '0';
	else if (c < 'A')
		return -1;
	else if (c <= 'F')
		return c - 'A' + 10;
	else if (c < 'a')
		return -1;
	else if (c <= 'f')
		return c - 'a' + 10;
	else
		return -1;
}

void searchfile(const char *filename, int fd, const unsigned char *value, const unsigned char *mask, int len, int errors, int solid)
{
	off_t offset = 0;
	unsigned char buf[pattern_size*4];



	while (1)
	{
		int r;

		memcpy(buf, buf + len, len);
		r = read(fd, buf + len, (pattern_size*4) - len);

		if (r < 0)
		{
			perror("read");
			return;
		} else if (!r)
			return;
		
		int o, i, err=0;


		for (o = offset ? len : 0; o < r - len + 1; ++o)
		{
			for (err=0,i = 0; i < len; ++i)
				if (((buf[o + i] != value[i]) && (mask[i] >2)) || ((buf[o + i] == 0) && (mask[i] == 1))) {
			// mask[i]==0 -> ?? (any value)
			// mask[i]==1 -> NZ (non-zero)

			// XXX performance tweak... put this into whole separate branch?
			// XXX or gcc -O3 do this automatically? 

					if (err++>errors) { //tolerate up to "errors" substitutions
						break;
					} else {
						// performance tweak for matching with errors
						// first few bytes can be required without errors
						if (i < solid) { break; }  
					}
				}
			if (i == len)
			{
				printf("%s: %08llx\n", filename, (unsigned long long)(offset + o - len));
			}
		}
		
		offset += r;
		
	}
}

void recurse(const char *path, const unsigned char *value, const unsigned char *mask, int len, int errors, int solid)
{
	struct stat s;
	if (stat(path, &s))
	{
		perror("stat");
		return;
	}
	if (!S_ISDIR(s.st_mode))
	{
		int fd = open(path, O_RDONLY);
		if (fd < 0)
			perror(path);
		else
		{
			searchfile(path, fd, value, mask, len, errors, solid);
			close(fd);
		}
		return;
	}

	DIR *dir = opendir(path);
	if (!dir)
	{
		perror(path);
		exit(3);
	}
	
	struct dirent *d;
	while ((d = readdir(dir)))
	{
		if (!(strcmp(d->d_name, ".") && strcmp(d->d_name, "..")))
			continue;
		char newpath[strlen(path) + strlen(d->d_name) + 1];
		strcpy(newpath, path);
		strcat(newpath, "/");
		strcat(newpath, d->d_name);
		recurse(newpath, value, mask, len, errors, solid);
	}
	
	closedir(dir);
}

int main(int argc, char **argv)
{
	unsigned char value[pattern_size], mask[pattern_size];
	int opt = 0, i = 0, len = 0, errors = 0, solid = 0;

	static struct option long_opts[]={
		{"err",1,0,'e'},
		{"help",0,0,'h'},
		{"solid",1,0,'s'},
		{0,0,0,0}
	};

	while((opt = getopt_long(argc, argv, "e:hs:",long_opts, &i)) != -1) {
		switch(opt) {
			case 'e': /*--err*/
				errors=atoi(optarg);
				break;
			case 'h': /*--help*/
				printf("usage: %s [<options>] <hex> [<path> [...]]\n",*argv);
				printf("--err <num>: tolerate up to num substitutions\n");
				printf("--help : reading now\n");
				printf("--solid <num> : Treat first num bytes as immutable\n");
				return 0;
				break;
			case 's': /*--solid*/
				solid=atoi(optarg);
				break;
			case '?': // err
				fprintf(stderr, "Syntax error.\n");
				return 7;
				break;
		}
	}

	if (argc < 2)
	{
		fprintf(stderr, "usage: %s [<options>] <hex> [<path> [...]]\n", *argv);
		return 1;
	}
	
	char *h = argv[optind++];
	while (*h && h[1] && len < pattern_size)
	{
		if (h[0] == '?' && h[1] == '?')
		{
			value[len] = mask[len] = 0;
			len++;
			h += 2;
		} else 
		{
			if (h[0] == 'N' && h[1] == 'Z') 
			{
				value[len] = 0; mask[len] = 1;
				len++;
				h += 2;
			} else
			{
				int v0 = ascii2hex(*h++);
				int v1 = ascii2hex(*h++);
				
				if ((v0 == -1) || (v1 == -1))
				{
					fprintf(stderr, "invalid hex string!\n");
					return 2;
				}
				value[len] = (v0 << 4) | v1; mask[len++] = 0xFF;
			}
		}
	}
	
	if (!len || *h)
	{
		fprintf(stderr, "invalid/empty search string\n");
		return 2;
	}
	
	if (argc < 3) {
		searchfile("stdin", 0, value, mask, len, errors, solid);
	} else {
		int c = optind;
		while (c < argc)
			recurse(argv[c++], value, mask, len, errors, solid);
	}
	return 0;
}
